/*
 * TCP protocol
 * Copyright (c) 2002 Fabrice Bellard
 *
 * This file is part of FFmpeg.
 *
 * FFmpeg is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * FFmpeg is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with FFmpeg; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA
 */
#include <android/log.h>
#include "avformat.h"
#include <unistd.h>
#include "internal.h"
#include "network.h"
#include "os_support.h"
#if HAVE_SYS_SELECT_H
#include <sys/select.h>
#endif
#include <sys/time.h>


//DK vvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvv
#include "shared_notify.h"

//added for bytes per second count
int giTcp_read_len = 0;
uint64_t gu64maxConnectTimeVal = TCP_DEFAULT_CONNECT_TIMEOUT_MS * 1000; //conver from ms to us
uint64_t gu64maxTcpRecvTimeVal = TCP_DEFAULT_RECV_TIMEOUT_MS * 1000; //conver from ms to us

static int default_tcp_recv_cb(void);

static ff_callback *tcp_recv_start_cb = default_tcp_recv_cb;
static ff_callback *tcp_recv_retry_cb = default_tcp_recv_cb;


void tcp_set_recv_start_cb(ff_callback *new_cb)
{
    if (!new_cb)
        new_cb = default_tcp_recv_cb;
    tcp_recv_start_cb = new_cb;
}

void tcp_set_recv_retry_cb(ff_callback *new_cb)
{
    if (!new_cb)
        new_cb = default_tcp_recv_cb;
    tcp_recv_retry_cb = new_cb;
}

static int default_tcp_recv_cb(void)
{
    //0 indicates "OK"
    return 0;
}

//DK ^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^


typedef struct TCPContext {
    int fd;
} TCPContext;

/* return non zero if error */
static int tcp_open(URLContext *h, const char *uri, int flags)
{
    struct addrinfo hints, *ai, *cur_ai;
    int port, fd = -1;
    TCPContext *s = NULL;
    fd_set wfds;
    int fd_max, ret;
    struct timeval tv;
    socklen_t optlen;
    char hostname[1024],proto[1024],path[1024];
    char portstr[10];
    uint64_t max_connect_time;

    __android_log_print ( ANDROID_LOG_INFO, "tcp_open", "enter. uri=%s\n", uri);

    ff_url_split(proto, sizeof(proto), NULL, 0, hostname, sizeof(hostname),
        &port, path, sizeof(path), uri);
    if (strcmp(proto,"tcp") || port <= 0 || port >= 65536)
        return AVERROR(EINVAL);

    memset(&hints, 0, sizeof(hints));
    hints.ai_family = AF_UNSPEC;
    hints.ai_socktype = SOCK_STREAM;
    snprintf(portstr, sizeof(portstr), "%d", port);
    if (getaddrinfo(hostname, portstr, &hints, &ai))
    {
        __android_log_print ( ANDROID_LOG_INFO, "tcp_open", "getaddrinfo failed\n");
        err_notify_cb(ID_NOTIFY_ERROR_TCP_GETADDRINFO_FAILED, VALUE_NOTIFY_ERROR_NO_EXTRA_INFO);
        return AVERROR(EIO);
    }

    cur_ai = ai;

    /*calculate the system time of when we'll exceed our max connect time */
    max_connect_time = av_gettime() + gu64maxConnectTimeVal;

 restart:
    fd = socket(cur_ai->ai_family, cur_ai->ai_socktype, cur_ai->ai_protocol);
    if (fd < 0)
    {
        __android_log_print ( ANDROID_LOG_INFO, "tcp_open", "socket failed.  fd=%08X\n", fd);
        err_notify_cb(ID_NOTIFY_ERROR_TCP_SOCKET_OPEN_FAILED, fd);
        goto fail;
    }
    ff_socket_nonblock(fd, 1);

 redo:
    ret = connect(fd, cur_ai->ai_addr, cur_ai->ai_addrlen);
    if (ret < 0) {
        if (ff_neterrno() == FF_NETERROR(EINTR))
            goto redo;
        if (ff_neterrno() != FF_NETERROR(EINPROGRESS) &&
            ff_neterrno() != FF_NETERROR(EAGAIN))
        {
            __android_log_print ( ANDROID_LOG_INFO, "tcp_open", "connect failed.  ff_neterrno()=%08X\n", ff_neterrno());
            err_notify_cb(ID_NOTIFY_ERROR_TCP_CONNECT_FAILED, ff_neterrno());
            goto fail;
        }

        /* wait until we are connected or until abort */
        for(;;) {
            if (url_interrupt_cb()) {
                ret = AVERROR(EINTR);
                err_notify_cb(ID_NOTIFY_ERROR_TCP_OPEN_ABORTED, VALUE_NOTIFY_ERROR_NO_EXTRA_INFO);
                goto fail1;
            }
            fd_max = fd;
            FD_ZERO(&wfds);
            FD_SET(fd, &wfds);
            tv.tv_sec = 0;
            tv.tv_usec = 100 * 1000;
            ret = select(fd_max + 1, NULL, &wfds, NULL, &tv);
            if (ret > 0 && FD_ISSET(fd, &wfds))
                break;

            //check to see if we've exceeded the max amount of time we're supposed to try for
            if (av_gettime() >= max_connect_time) {
                ret = AVERROR(EINTR);
                err_notify_cb(ID_NOTIFY_ERROR_TCP_OPEN_TIMED_OUT, VALUE_NOTIFY_ERROR_NO_EXTRA_INFO);
                goto fail1;
            }
        }

        /* test error */
        optlen = sizeof(ret);
        getsockopt (fd, SOL_SOCKET, SO_ERROR, &ret, &optlen);
        if (ret != 0)
        {
            err_notify_cb(ID_NOTIFY_ERROR_TCP_SOCKET_OPTIONS_FAILED, ret);
            goto fail;
        }
    }
    s = av_malloc(sizeof(TCPContext));
    if (!s) {
        freeaddrinfo(ai);
        return AVERROR(ENOMEM);
    }
    h->priv_data = s;
    h->is_streamed = 1;
    s->fd = fd;
    freeaddrinfo(ai);
    return 0;

 fail:

    if (cur_ai->ai_next) {
        /* Retry with the next sockaddr */
        cur_ai = cur_ai->ai_next;
        if (fd >= 0)
            closesocket(fd);
        goto restart;
    }
    __android_log_print ( ANDROID_LOG_INFO, "tcp_open", "FAIL\n");
    ret = AVERROR(EIO);
 fail1:
    if (fd >= 0)
        closesocket(fd);
    freeaddrinfo(ai);
    __android_log_print ( ANDROID_LOG_INFO, "tcp_open", "FAIL1\n");
    return ret;
}



static int tcp_read(URLContext *h, uint8_t *buf, int size)
{
    TCPContext *s = h->priv_data;
    int len, fd_max, ret;
    fd_set rfds;
    struct timeval tv;
    
    tcp_recv_start_cb();

    for (;;) {

//      __android_log_print ( ANDROID_LOG_INFO, "tcp_read", "size=%d\n", size);

        if (url_interrupt_cb())
        {
            err_notify_cb(ID_NOTIFY_ERROR_TCP_READ_ABORTED, VALUE_NOTIFY_ERROR_NO_EXTRA_INFO);
            return AVERROR(EINTR);
        }

        //__android_log_print ( ANDROID_LOG_INFO, "tcp_read", "after url_interrupt_cb\n");
        fd_max = s->fd;
        FD_ZERO(&rfds);
        FD_SET(s->fd, &rfds);
        tv.tv_sec = 0;
        tv.tv_usec = 100 * 1000;
        ret = select(fd_max + 1, &rfds, NULL, NULL, &tv);
        if (ret > 0 && FD_ISSET(s->fd, &rfds)) 
        {
            len = recv(s->fd, buf, size, 0);
            if (len < 0) {
                if (ff_neterrno() != FF_NETERROR(EINTR) &&
                    ff_neterrno() != FF_NETERROR(EAGAIN))
                {   
                    err_notify_cb(ID_NOTIFY_ERROR_TCP_READ_RECEIVE_FAILED, ff_neterrno()); 
                    return AVERROR(ff_neterrno());
                }
            } 
            else 
            {    
                giTcp_read_len += len;
                return len;
            }
        } 
        else if (ret < 0) 
        {
            if (ff_neterrno() == FF_NETERROR(EINTR))
                continue;
            err_notify_cb(ID_NOTIFY_ERROR_TCP_READ_WAIT_FAILED, ff_neterrno()); 
            return -1;
        }
        
        //if callback returns 1 then we've timedout
        if (tcp_recv_retry_cb())
        {
            err_notify_cb(ID_NOTIFY_ERROR_TCP_READ_RECV_TIMED_OUT, VALUE_NOTIFY_ERROR_NO_EXTRA_INFO);
            return AVERROR(ETIMEDOUT);
        }
    }
}

static int tcp_write(URLContext *h, uint8_t *buf, int size)
{
    TCPContext *s = h->priv_data;
    int ret, size1, fd_max, len;
    fd_set wfds;
    struct timeval tv;

    size1 = size;
    while (size > 0) {
        if (url_interrupt_cb())
        {
            err_notify_cb(ID_NOTIFY_ERROR_TCP_WRITE_ABORTED, VALUE_NOTIFY_ERROR_NO_EXTRA_INFO);
            return AVERROR(EINTR);
        }
        fd_max = s->fd;
        FD_ZERO(&wfds);
        FD_SET(s->fd, &wfds);
        tv.tv_sec = 0;
        tv.tv_usec = 100 * 1000;
        ret = select(fd_max + 1, NULL, &wfds, NULL, &tv);
        if (ret > 0 && FD_ISSET(s->fd, &wfds)) {
            len = send(s->fd, buf, size, 0);
            if (len < 0) {
                if (ff_neterrno() != FF_NETERROR(EINTR) &&
                    ff_neterrno() != FF_NETERROR(EAGAIN))
                {    
                    err_notify_cb(ID_NOTIFY_ERROR_TCP_WRITE_SEND_FAILED, ff_neterrno());
                    return AVERROR(ff_neterrno());
                }
                continue;
            }
            size -= len;
            buf += len;
        } else if (ret < 0) {
            if (ff_neterrno() == FF_NETERROR(EINTR))
                continue;
            err_notify_cb(ID_NOTIFY_ERROR_TCP_WRITE_WAIT_FAILED, ff_neterrno());
            return -1;
        }
    }
    return size1 - size;
}

static int tcp_close(URLContext *h)
{
    TCPContext *s = h->priv_data;
    closesocket(s->fd);
    av_free(s);
    return 0;
}

static int tcp_get_file_handle(URLContext *h)
{
    TCPContext *s = h->priv_data;
    return s->fd;
}

URLProtocol tcp_protocol = {
    "tcp",
    tcp_open,
    tcp_read,
    tcp_write,
    NULL, /* seek */
    tcp_close,
    .url_get_file_handle = tcp_get_file_handle,
};
